from .flat_kmers import letter_sequence_to_numeric, numeric_to_letter_sequence
from .snp_kmer_finder import kmer_to_hash_fast, sequence_to_kmer_hash, kmer_hash_to_sequence
from .snp_kmer_finder import SnpKmerFinder
from .flat_kmers import FlatKmers
from .reverse_kmer_index import ReverseKmerIndex
from .collision_free_kmer_index import CollisionFreeKmerIndex
from .collision_free_kmer_index import CollisionFreeKmerIndex as KmerIndex
from .unique_kmer_index import UniqueKmerIndex
from .reference_kmer_index import ReferenceKmerIndex
from .read_kmers import ReadKmers



from .snp_kmer_finder import SnpKmerFinder

class IndelKmerFinder(SnpKmerFinder):


    def __init__(self, graph, k=15, spacing=None, include_reverse_complements=False, pruning=False, max_kmers_same_position=100000,
                 max_frequency=10000, max_variant_nodes=10000, only_add_variant_kmers=False, whitelist=None, only_save_variant_nodes=False,
                 start_position=None, end_position=None):

        super(IndelKmerFinder, self).__init__(graph, k, spacing, include_reverse_complements, pruning, max_kmers_same_position,
                 max_frequency, max_variant_nodes, only_add_variant_kmers, whitelist, only_save_variant_nodes,
                 start_position, end_position)





import logging
import numpy as np


class ReverseKmerIndex:
    properties = {"nodes_to_index_positions", "nodes_to_n_hashes", "hashes", "ref_positions"}
    def __init__(self, nodes_to_index_positions=None, nodes_to_n_hashes=None, hashes=None, ref_positions=None):
        self.nodes_to_index_positions = nodes_to_index_positions
        self.nodes_to_n_hashes = nodes_to_n_hashes
        self.hashes = hashes
        self.ref_positions = ref_positions

    def __str__(self):
        description = "Nodes to index positions: %s\n" % self.nodes_to_index_positions
        description += "Nodes to n hashes      : %s\n" % self.nodes_to_n_hashes
        description += "Hashes:                  %s\n" % self.hashes
        description += "Ref positions:                  %s\n" % self.ref_positions
        return description

    def __repr_(self):
        return self.__str__()

    def get_node_kmers(self, node):
        index_position = int(self.nodes_to_index_positions[node])
        n_hashes = int(self.nodes_to_n_hashes[node])
        if n_hashes == 0:
            return []

        return self.hashes[index_position:index_position+n_hashes]

    def get_node_kmers_and_ref_positions(self, node):
        try:
            index_position = int(self.nodes_to_index_positions[node])
        except IndexError:
            logging.error("Invalid node %d" % node)
            raise

        n_hashes = int(self.nodes_to_n_hashes[node])
        if n_hashes == 0:
            return [[], []]

        return self.hashes[index_position:index_position+n_hashes], self.ref_positions[index_position:index_position+n_hashes]

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name)
        except FileNotFoundError:
            data = np.load(file_name + ".npz")

        return cls(data["nodes_to_index_positions"], data["nodes_to_n_hashes"], data["hashes"], data["ref_positions"])

    def to_file(self, file_name):
        np.savez(file_name, nodes_to_index_positions=self.nodes_to_index_positions,
                            nodes_to_n_hashes=self.nodes_to_n_hashes,
                            hashes=self.hashes,
                            ref_positions=self.ref_positions)

    @classmethod
    def from_flat_kmers(cls, flat_kmers):
        logging.info("Creating ReverseKmerIndex from flat kmers")
        nodes = flat_kmers._nodes
        kmers = flat_kmers._hashes
        ref_positions = flat_kmers._ref_offsets

        max_node = np.max(nodes)
        logging.info("Max node: %d" % max_node)

        nodes_index = np.zeros(max_node+1, dtype=np.uint32)
        n_kmers = np.zeros(max_node+1, dtype=np.uint16)
        sorted_nodes = np.argsort(nodes)
        nodes = nodes[sorted_nodes]
        kmers = kmers[sorted_nodes]
        ref_positions = ref_positions[sorted_nodes]


        diffs = np.ediff1d(nodes, to_begin=1)
        positions_of_unique_nodes = np.nonzero(diffs)[0]
        unique_nodes = nodes[positions_of_unique_nodes]

        nodes_index[unique_nodes] = positions_of_unique_nodes
        n_kmers_numbers = np.ediff1d(positions_of_unique_nodes, to_end=len(nodes)-positions_of_unique_nodes[-1])
        n_kmers[unique_nodes] = n_kmers_numbers
        return cls(nodes_index, n_kmers, kmers, ref_positions)





import logging
import numpy as np
from collections import defaultdict


class FlatKmers:
    def __init__(self, hashes, nodes, ref_offsets=None, allele_frequencies=None):
        assert len(hashes) == len(nodes)
        self._hashes = hashes
        self._nodes = nodes
        if ref_offsets is None:
            self._ref_offsets = np.zeros(len(self._nodes))
        else:
            self._ref_offsets = ref_offsets

        if allele_frequencies is None:
            logging.info("Allele frequencies not provided. Setting all to 1.0")
            self._allele_frequencies = np.zeros(len(self._hashes), dtype=np.single) + 1.0
        else:
            self._allele_frequencies = allele_frequencies


    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name)
        except FileNotFoundError:
            data = np.load(file_name + ".npz")

        logging.info("Loaded kmers from %s" % file_name)
        return cls(data["hashes"], data["nodes"], data["ref_offsets"], data["allele_frequencies"])

    def to_file(self, file_name):
        np.savez(file_name, hashes=self._hashes, nodes=self._nodes, ref_offsets=self._ref_offsets, allele_frequencies=self._allele_frequencies)
        logging.info("Save dto %s.npz" % file_name)


    @classmethod
    def from_multiple_flat_kmers(cls, flat_kmers_list):
        logging.info("Making flat kmers")
        hashes = []
        nodes = []
        ref_offsets = []
        allele_frequencies = []
        for flat in flat_kmers_list:
            hashes.extend(flat._hashes)
            nodes.extend(flat._nodes)
            if flat._ref_offsets is not None:
                ref_offsets.extend(flat._ref_offsets)

            allele_frequencies.extend(flat._allele_frequencies)

        if len(ref_offsets) == 0:
            ref_offsets = None
        else:
            ref_offsets = np.array(ref_offsets, np.uint64)

        logging.info("Done making flat kmers")
        return FlatKmers(np.array(hashes, dtype=np.uint64), np.array(nodes, np.uint32), ref_offsets, np.array(allele_frequencies, dtype=np.uint16))

    def sum_of_kmer_frequencies(self, kmer_index_with_frequencies):
        return sum([0] + [max(1, kmer_index_with_frequencies.get_frequency(int(kmer))) for kmer in self._hashes])

    def maximum_kmer_frequency(self, kmer_index_with_frequencies):
        return max([0] + [kmer_index_with_frequencies.get_frequency(int(kmer)) for kmer in self._hashes])
    
    def get_new_without_singletons(self):
        has_been_traversed = set()
        new_hashes = []
        new_nodes = []
        new_ref_offsets = []
        new_allele_frequencies = []

        for i in range(len(self._hashes)):
            if i % 1000000 == 0:
                logging.info("%d entries processed. %d kept" % (i, len(new_hashes)))

            hash = self._hashes[i]
            if hash in has_been_traversed:
                new_hashes.append(hash)
                new_nodes.append(self._nodes[i])
                new_ref_offsets.append(self._ref_offsets[i])
                new_allele_frequencies.append(self._allele_frequencies[i])
            else:
                has_been_traversed.add(hash)

        logging.info("Making new numpy arrays")
        new_hashes = np.array(new_hashes, dtype=self._hashes.dtype)
        new_nodes = np.array(new_nodes, dtype=self._nodes.dtype)
        new_ref_offsets = np.array(new_ref_offsets, dtype=self._ref_offsets.dtype)
        new_allele_frequencies = np.array(new_allele_frequencies, dtype=self._allele_frequencies.dtype)

        logging.info("Returning new flat kmers")
        return FlatKmers(new_hashes, new_nodes, new_ref_offsets, new_allele_frequencies)


def letter_sequence_to_numeric(sequence):
    if not isinstance(sequence, np.ndarray):
        sequence = np.array(list(sequence.lower()), dtype="<U1")

    numeric = np.zeros_like(sequence, dtype=np.int64)
    numeric[np.where(sequence == "n")[0]] = 0
    numeric[np.where(sequence == "a")[0]] = 0
    numeric[np.where(sequence == "c")[0]] = 1
    numeric[np.where(sequence == "t")[0]] = 2
    numeric[np.where(sequence == "g")[0]] = 3
    numeric[np.where(sequence == "m")[0]] = 0
    return numeric

def numeric_to_letter_sequence(sequence):

    numeric = np.zeros_like(sequence, dtype=object)
    numeric[np.where(sequence == 0)[0]] = "a"
    numeric[np.where(sequence == 1)[0]] = "c"
    numeric[np.where(sequence == 2)[0]] = "t"
    numeric[np.where(sequence == 3)[0]] = "g"
    return numeric








import random
import pickle
import time

import SharedArray as sa
import logging
import numpy as np
from itertools import repeat
from pathos.multiprocessing import Pool

class SingleSharedArray:
    properties = {"array"}
    def __init__(self, array=None):
        self.array = array

SHARED_MEMORIES_IN_SESSION = []

_shared_pool = None


def get_shared_pool(n_threads=16):
    global _shared_pool
    if _shared_pool is None:
        _shared_pool = Pool(n_threads)
        logging.info("Made shared pool")

    return _shared_pool


def close_shared_pool():
    global _shared_pool
    if _shared_pool is not None:
        _shared_pool.close()
        logging.info("Closed shared pool")


def to_shared_memory(object, name):
    global SHARED_MEMORIES_IN_SESSION
    t = time.perf_counter()
    logging.debug("Writing to shared memory %s" % name)
    meta_information = {}
    for property_name in object.properties:
        data = object.__getattribute__(property_name)

        if data is None:
            data = np.zeros(0)

        # Wrap single ints in arrays
        if data.shape == ():
            data = np.array([data], dtype=data.dtype)

        data_type = data.dtype
        data_shape = data.shape
        meta_information[property_name] = (data_type, data_shape)

        # Make shared memory and copy data to buffer
        #logging.info("Field %s has shape %s and type %s" % (property_name, data_shape, data_type))
        array_name = name + "__" + property_name
        try:
            sa.delete(array_name)
        except FileNotFoundError:
            logging.debug("No existing shared memory, can create new one")

        shared_array = sa.create(array_name, data_shape, data_type)
        shared_array[:] = data
        SHARED_MEMORIES_IN_SESSION.append(array_name)

    f = open(name + "_meta.shm", "wb")
    pickle.dump(meta_information, f)
    #logging.info("Done writing to shared memory. Took %.3f sec" % (time.perf_counter()-t))
    #logging.info("Wrote meta data to file")


def from_shared_memory(cls, name):
    object = cls()
    t = time.perf_counter()
    try:
        meta_data = pickle.load(open(name + "_meta.shm", "rb"))
    except EOFError:
        logging.error("Failed reading from %s._meta.shm" % name)
        raise

    for property_name, data in meta_data.items():
        data_type = data[0]
        data_shape = data[1]
        #logging.info("Found property %s with shape %s and type %s" % (property_name, data_shape, data_type))
        data = sa.attach(name + "__" + property_name)
        # Single ints are wrapped in arrays
        if len(data) == 1 and property_name == "_modulo":
            data = data[0]
            #logging.info("Extracted single int from %s" % property_name)
        setattr(object, property_name, data)
        #print("Data sample from class:: %s" % object.__getattribute__(property_name))


    return object


def remove_shared_memory(name):
    shared_memories = [s.name.decode("utf-8") for s in sa.list()]

    for m in shared_memories:
        if m.startswith(name + "__"):
            sa.delete(m)
            return

    logging.warning("No shared memory with name %s" % name)
    logging.warning("Available shared memories: %s" % shared_memories)


def remove_shared_memory_in_session():
    for name in SHARED_MEMORIES_IN_SESSION:
        try:
            sa.delete(name)
        except FileNotFoundError:
            logging.warning("Tried to deleted shared memory %s that did not exist" % name)

def remove_all_shared_memory():
    for shared in sa.list():
        sa.delete(shared.name.decode("utf-8"))


def _run_numpy_based_function_on_shared_memory_arguments(function, arguments, interval):
    start_time = time.perf_counter()
    start, end = interval
    #logging.info("Running on interval %d-%d" % (start, end))
    arguments = [from_shared_memory(SingleSharedArray, a).array if type(a) == str else a for a in arguments]
    sliced_arguments = []
    for argument in arguments:
        if isinstance(argument, np.ndarray):
            argument = argument[start:end]
        sliced_arguments.append(argument)

    result = function(*sliced_arguments)
    shared_memory_name = str(random.randint(0,10e15))
    to_shared_memory(SingleSharedArray(result), shared_memory_name)
    #logging.info("Interval %d-%d took %.4f sec" % (start, end, time.perf_counter()-start_time))
    return shared_memory_name


def run_numpy_based_function_in_parallel(function, n_threads, arguments):
    # split every argument that is not an int
    # put each in shared memory
    # run function on chunk on separete thread, put in result array
    # return result array
    new_arguments = []

    # Put np arrays in shared memory, everything else we keep as is
    array_length = 0
    for argument in arguments:
        if isinstance(argument, np.ndarray):
            argument_id = str(np.random.randint(0, 10e15))
            to_shared_memory(SingleSharedArray(argument), argument_id)
            new_arguments.append(argument_id)
            assert array_length == 0 or array_length == argument.shape[0], "Found argument with different shape"
            array_length = argument.shape[0]

        else:
            new_arguments.append(argument)

    t = time.perf_counter()
    pool = get_shared_pool(n_threads)  # Pool(n_threads)

    intervals = list([int(i) for i in np.linspace(0, array_length, n_threads + 1)])
    intervals = [(from_pos, to_pos) for from_pos, to_pos in zip(intervals[0:-1], intervals[1:])]
    #logging.info("Will run on intervals %s" % intervals)

    results = []
    for result in pool.starmap(_run_numpy_based_function_on_shared_memory_arguments, zip(repeat(function), repeat(new_arguments), intervals)):
        result = from_shared_memory(SingleSharedArray, result).array
        results.append(result)

    t = time.perf_counter()
    results = np.concatenate(results)
    logging.info("Time to concatenate results: %.3f" % (time.perf_counter()-t))
    return results







import logging
import time

from .snp_kmer_finder import SnpKmerFinder
from .flat_kmers import FlatKmers
from obgraph import VariantNotFoundException


class UniqueVariantKmersFinder:
    def __init__(self, graph, variant_to_nodes, variants, k=31, max_variant_nodes=6, kmer_index_with_frequencies=None, haplotype_matrix=None, node_to_variants=None,
                 do_not_choose_lowest_frequency_kmers=False):
        self.graph = graph
        self.variant_to_nodes = variant_to_nodes
        self.reference_kmer_index = None
        self.variants = variants
        self.k = k
        self.flat_kmers_found = []
        self.n_failed_variants = 0
        self._n_skipped_because_added_on_other_node = 0
        self._max_variant_nodes = max_variant_nodes
        self._kmer_index_with_frequencies = kmer_index_with_frequencies
        self.haplotype_matrix = haplotype_matrix
        self.node_to_variants = node_to_variants

        self._choose_kmers_with_lowest_frequencies = True
        if do_not_choose_lowest_frequency_kmers:
            self._choose_kmers_with_lowest_frequencies = False

    def kmer_is_unique_on_reference_position(self, kmer, reference_position, ref_start, ref_end):
        # returns true if kmer is unique when ignoring kmers on same ref pos
        reference_kmers = self.reference_kmer_index.get_between(ref_start, ref_end)
        # Remove the one matching our position
        #logging.info("    Checking whether %d among reference kmers: %s" % (kmer, reference_kmers))
        for i, reference_kmer in enumerate(reference_kmers):
            pos = ref_start + i
            if pos != reference_position and reference_kmer == kmer:
                return False
        return True

    def find_unique_kmers_over_variant(self, variant, ref_node, variant_node):
        is_valid = False
        has_added = False
        # Start searching from before. The last position is probably the best for indels, since the sequence after is offset-ed.
        # We want to end on this, because this is what is chosen if all else fails
        possible_ref_positions = [variant.position - i for i in range(1, self.k - 2)][::-1]
        valid_positions_found = []

        for possible_ref_position in possible_ref_positions:
            possible_ref_position_adjusted = self.graph.convert_chromosome_ref_offset_to_graph_ref_offset(possible_ref_position, variant.chromosome)
            is_valid = True
            finder = SnpKmerFinder(self.graph, self.k, max_variant_nodes=self._max_variant_nodes, only_store_nodes=set([ref_node, variant_node]),
                                   haplotype_matrix=self.haplotype_matrix, node_to_variants=self.node_to_variants, variant_to_nodes=self.variant_to_nodes)
            finder.find_kmers_from_linear_ref_position(possible_ref_position_adjusted)


            kmers_ref = set()
            kmers_variant = set()
            for kmer, nodes, ref_position, hash in finder.kmers_found:
                if ref_node in nodes:
                    kmers_ref.add(kmer.lower())
                if variant_node in nodes:
                    kmers_variant.add(kmer.lower())

                # not a requirement anymore
                #if not self.kmer_is_unique_on_reference_position(hash, ref_position, max(0, ref_position - 150), ref_position + 150 - self.k):
                #    is_valid = False
                #    #logging.info("  Found on linear ref elsewhere, %s" % kmer)
                #    break

                #if hash in self._hashes_added:
                #    is_valid = False
                #    self._n_skipped_because_added_on_other_node += 1

            if not is_valid:
                continue

            # Check if there are identical kmers on the two variant nodes
            if len(kmers_ref.intersection(kmers_variant)) > 0:
                #logging.info("    Felles kmer mellom ref og variant: %s / %s" % (kmers_ref, kmers_variant))
                is_valid = False

            # if we are at last position (then choose this anyway, better than nothing)
            if possible_ref_position == possible_ref_positions[-1]:
                is_valid = True
                #is_valid = False  # testing this

            if is_valid:
                # Kmers are valid, we don't need to search anymore for this variant
                flat = finder.get_flat_kmers()
                valid_positions_found.append(flat)

                if flat.maximum_kmer_frequency(self._kmer_index_with_frequencies) <= 1:
                    # no need to search for more positions, all kemrs are unique
                    break

                if len(flat._nodes) == 0:
                    logging.warning("Found 0 nodes for variant %s. Hashes: %s, ref positions: %s. Searched from ref position %d" % (variant, flat._hashes, flat._ref_offsets, possible_ref_position))
                    #raise Exception("No kmers found")

                if ref_node not in flat._nodes:
                    logging.warning("Found kmers for variant %s with ref/variant nodes %d/%d but flat kmers does not contain ref node. Flat kmer nodes: %s. Searched from possition %d" % (variant, ref_node, variant_node, flat._nodes, possible_ref_position))
                    #raise Exception("No kmers found")

                if variant_node not in flat._nodes:
                    logging.warning("No variant node kmers found for variant %s with variant node %d and ref node %d" % (variant, variant_node, ref_node))
                    logging.warning("Found no variant node kmers for variant %s. Hashes: %s, ref positions: %s. Searched from ref position %d" % (variant, flat._hashes, flat._ref_offsets, possible_ref_position))

        # Sort positions by max kmer frequency
        if len(valid_positions_found) == 0:
            logging.warning("Found no positions with valid kmers for variant %s" % variant)
            self.n_failed_variants += 1
            return

        if self._choose_kmers_with_lowest_frequencies:
            valid_positions_found = sorted(valid_positions_found, key=lambda p: p.maximum_kmer_frequency(self._kmer_index_with_frequencies))

        #valid_positions_found = sorted(valid_positions_found, key=lambda p: p.sum_of_kmer_frequencies(self._kmer_index_with_frequencies))
        best_position = valid_positions_found[0]
        self.flat_kmers_found.append(best_position)


    def find_unique_kmers(self):
        prev_time = time.time()
        for i, variant in enumerate(self.variants):
            n_processed = len(self.flat_kmers_found)
            if i % 10000 == 0:
                logging.info("%d/%d variants processed (time spent on previous 10000 variants: %.3f s). "
                             "Now on chromosome/ref pos %s/%d" % (i, len(self.variants), time.time()-prev_time, str(variant.chromosome), variant.position))
                prev_time = time.time()

            #ref_node, variant_node = self.graph.get_variant_nodes(variant)
            ref_node = self.variant_to_nodes.ref_nodes[variant.vcf_line_number]
            variant_node = self.variant_to_nodes.var_nodes[variant.vcf_line_number]

            if ref_node == 0 or variant_node == 0:
                continue

            self.find_unique_kmers_over_variant(variant, ref_node, variant_node)

            if len(self.flat_kmers_found) != n_processed + 1:
                logging.warning("DID NOT FIND KMERS ON %s" % variant)

        logging.info("N variants with kmers found: %d" % len(self.flat_kmers_found))
        logging.info("Done with all variants. N that failed: %d" % self.n_failed_variants)

        return FlatKmers.from_multiple_flat_kmers(self.flat_kmers_found)




import numpy as np
import logging
import pickle

from graph_kmer_index import kmer_hash_to_sequence, sequence_to_kmer_hash
from Bio.Seq import Seq

class CollisionFreeKmerIndex:
    properties = {
            "_hashes_to_index",
            "_n_kmers",
            "_nodes",
            "_ref_offsets",
            "_kmers",
            "_modulo",
            "_frequencies",
            "_allele_frequencies"
        }

    def __init__(self, hashes_to_index=None, n_kmers=None, nodes=None, ref_offsets=None, kmers=None, modulo=452930477, frequencies=None, allele_frequencies=None):
        self._hashes_to_index = hashes_to_index
        self._n_kmers = n_kmers
        self._nodes = nodes
        self._ref_offsets = ref_offsets
        self._kmers = kmers  # Actual numeric kmers (not hashes of numeric kmers) at each position
                             # used to filter out collisions
        self._modulo = modulo
        if frequencies is None:
            self._frequencies = 0
        else:
            self._frequencies = frequencies

        self._allele_frequencies = allele_frequencies

    def set_allele_frequencies(self, frequencies):
        pass

    def max_node_id(self):
        return np.max(self._nodes)

    def set_frequencies_using_other_index(self, other, multiplier=1, min_frequency=1):
        unique = np.unique(self._kmers)
        for i, kmer in enumerate(unique):
            kmer = int(kmer)
            if i % 100000 == 0:
                logging.info("%d/%d unique kmers processed" % (i, len(unique)))
            frequency = other.get_frequency(kmer)
            hash = int(kmer) % self._modulo
            position = self._hashes_to_index[hash]
            n_hits = self._n_kmers[hash]
            start = position
            end = position + n_hits
            hit_positions = np.where(self._kmers[start:end] == kmer)[0]
            self._frequencies[hit_positions + start] = max(min_frequency, frequency * multiplier)

    def set_frequencies(self, skip=False):
        logging.info("Setting frequencies")
        # Count number of each kmer (not hashes, store these)
        self._frequencies = np.zeros(len(self._kmers), dtype=np.uint16)

        if skip:
            logging.info("Skipped setting frequencies. All frequencies are just 0 by default.")
            return

        unique = np.unique(self._kmers)
        for i, kmer in enumerate(unique):
            if i % 100000 == 0:
                logging.info("%d/%d unique kmers processed" % (i, len(unique)))

            hash = int(kmer) % self._modulo
            position = self._hashes_to_index[hash]
            n_hits = self._n_kmers[hash]
            start = position
            #assert start != 0 or hash == 0, "Kmer %d with hash %d, index position %d not found in index" % (kmer, hash, position)
            end = position + n_hits
            hit_positions = np.where(self._kmers[start:end] == kmer)[0]

            # The count is the number of unique ref positions here
            # (since same entry can have multiple nodes, but always same ref pos)
            count = len(set(self._ref_offsets[hit_positions + start]))
            assert count > 0, "Count is not > 0 for kmer %d, start, end: %d,%d. Ref offsets: %s" % (kmer, start, end, self._ref_offsets[hit_positions + start])
            self._frequencies[hit_positions + start] = count

    def get(self, kmer, max_hits=10):
        hash = kmer % self._modulo
        position = self._hashes_to_index[hash]
        n_hits = self._n_kmers[hash]
        start = position
        end = position + n_hits
        hit_positions = np.where(self._kmers[start:end] == kmer)[0]
        frequencies = self._frequencies[hit_positions+start]
        allele_frequencies = self._allele_frequencies[hit_positions+start]
        if len(hit_positions) == 0 or frequencies[0] > max_hits:
            return None, None, None, None

        return self._nodes[hit_positions + start], self._ref_offsets[hit_positions + start], frequencies, allele_frequencies

    def get_grouped_nodes(self, kmer, max_hits=10):
        hits = self.get(kmer, max_hits)
        if hits[0] is None:
            return None

        ref_offsets = hits[1]
        nodes = hits[0]
        sorting = np.argsort(ref_offsets)
        ref_offsets = ref_offsets[sorting]
        nodes = nodes[sorting]


        _, hit_indexes = np.unique(ref_offsets, return_index=True)
        hit_indexes = list(hit_indexes)
        hit_indexes.append(len(ref_offsets))

        intervals = [(start, end) for start, end in zip(hit_indexes[0:-1], hit_indexes[1:])]
        return [nodes[start:end] for start, end in intervals]

    def get_frequency(self, kmer, include_reverse_complement=True, k=31):
        nodes, ref_offsets, frequencies, allele_frequencies = self.get(kmer, max_hits=1000000000000000)
        if nodes is None:
            f = 0
        else:
            f = int(frequencies[0])  # convert to avoid overflow error

        if include_reverse_complement:
            sequence = kmer_hash_to_sequence(kmer, k)
            rev_sequence = str(Seq(sequence).reverse_complement())
            rev_kmer = sequence_to_kmer_hash(rev_sequence)
            nodes, ref_offsets, frequencies, allele_frequencies = self.get(rev_kmer, max_hits=1000000000000000)

            if nodes is not None:
                f += int(frequencies[0])

        return f

    def get_nodes_and_ref_offsets_from_multiple_kmers(self, kmers, max_hits=10):
        all_nodes = []
        all_ref_offsets = []
        all_read_offsets = []
        all_frequencies = []
        for i, hash in enumerate(kmers):
            nodes, ref_offsets, frequencies, allele_frequencies = self.get(hash, max_hits=max_hits)
            if nodes is None:
                continue
            all_nodes.append(nodes)
            all_ref_offsets.append(ref_offsets)
            all_read_offsets.append(np.zeros(len(nodes)) + i)
            all_frequencies.append(frequencies)


        if len(all_nodes) == 0:
            return np.array([]), np.array([]), np.array([]), np.array([])

        all_nodes = np.concatenate(all_nodes)
        all_ref_offsets = np.concatenate(all_ref_offsets)
        all_read_offsets = np.concatenate(all_read_offsets)
        all_frequencies = np.concatenate(all_frequencies)
        return all_nodes, all_ref_offsets, all_read_offsets, all_frequencies

    def get_nodes_from_multiple_kmers(self, kmers, max_hits=10):
        all_nodes = []
        for i, hash in enumerate(kmers):
            nodes, ref_offsets, frequencies, allele_frequencies = self.get(hash, max_hits=max_hits)
            if nodes is None:
                continue
            all_nodes.append(nodes)


        if len(all_nodes) == 0:
            return np.array([])

        all_nodes = np.concatenate(all_nodes)
        return all_nodes

    def to_file(self, file_name):
        logging.info("Writing kmer index to file: %s" % file_name)
        np.savez(file_name, hashes_to_index=self._hashes_to_index,
                 n_kmers=self._n_kmers,
                 nodes=self._nodes,
                 ref_offsets=self._ref_offsets,
                 kmers=self._kmers,
                 modulo=self._modulo,
                 frequencies=self._frequencies,
                 allele_frequencies=self._allele_frequencies)

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name + ".npz")
        except FileNotFoundError:
            data = np.load(file_name)

        if "allele_frequencies" in data:
            allele_frequencies = data["allele_frequencies"]
        else:
            allele_frequencies = np.zeros(len(data["ref_offsets"]))

        return cls(data["hashes_to_index"], data["n_kmers"], data["nodes"], data["ref_offsets"], data["kmers"], data["modulo"], data["frequencies"], allele_frequencies)

    @classmethod
    def from_flat_kmers(cls, flat_kmers, modulo=452930477, skip_frequencies=False, skip_singletons=False):
        if skip_singletons:
            flat_kmers = flat_kmers.get_new_without_singletons()

        kmers = flat_kmers._hashes
        nodes = flat_kmers._nodes
        ref_offsets = flat_kmers._ref_offsets

        logging.info("Making hashes")
        hashes = kmers % modulo
        logging.info("Sorting")
        sorting = np.argsort(hashes)
        hashes = hashes[sorting]
        kmers = kmers[sorting]
        nodes = nodes[sorting]
        ref_offsets = ref_offsets[sorting]
        allele_frequencies = flat_kmers._allele_frequencies[sorting]
        logging.info("Done sorting")

        # Find positions where hashes change (these are our index entries)
        diffs = np.ediff1d(hashes, to_begin=1)
        unique_entry_positions = np.nonzero(diffs)[0]
        try:
            unique_hashes = hashes[unique_entry_positions]
        except IndexError:
            logging.info("unique entry positions: %s" % unique_entry_positions)
            logging.info("Hashes: %s" % hashes)
            raise

        lookup = np.zeros(modulo, dtype=np.int)
        lookup[unique_hashes] = unique_entry_positions
        n_entries = np.ediff1d(unique_entry_positions, to_end=len(nodes)-unique_entry_positions[-1])
        n_kmers = np.zeros(modulo, dtype=np.uint32)
        n_kmers[unique_hashes] = n_entries

        # Find out how many entries there are for each unique hash
        object = cls(lookup, n_kmers, nodes, ref_offsets, kmers, modulo, allele_frequencies=allele_frequencies)
        object.set_frequencies(skip_frequencies)

        if skip_singletons:
            logging.info("Adding 1 to all frequencies since singletons are skipped")
            object._frequencies += 1

        return object








import numpy as np
import logging


class KmerFrequencyIndex:
    def __init__(self, kmers, frequencies):
        self._kmers = kmers
        self._frequencies = frequencies

    def get(self, kmer):
        index = np.searchsorted(self._kmers, kmer, side="right")
        if self._kmers[index] == kmer:
            return self._frequencies[index]

        logging.warning("No hit for kmer %d" % kmer)
        return 0

    @classmethod
    def from_kmers(cls, kmers):
        logging.info("Sorting")
        sorting = np.argsort(kmers)
        kmers = kmers[sorting]
        logging.info("Counting")
        unique, frequencies = np.unique(kmers, return_counts=True)
        return cls(unique, frequencies)

    def to_file(self, file_name):
        np.savez(file_name, kmers=self._kmers, frequencies=self._frequencies)

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name)
        except FileNotFoundError:
            data = np.load(file_name + ".npz")

        return cls(data["kmers"], data["frequencies"])


from distutils.core import setup, Extension



setup(name='graph_kmer_index',
      version='0.0.1',
      description='Graph Kmer Index',
      url='http://github.com/ivargr/graph_kmer_index',
      author='Ivar Grytten',
      author_email='',
      license='MIT',
      packages=["graph_kmer_index"],
      zip_safe=False,
      install_requires=['numpy', 'sortedcontainers', 'tqdm', 'biopython', 'numba', 'cython', 'SharedArray'],
      classifiers=[
            'Programming Language :: Python :: 3'
      ],
      entry_points={
            'console_scripts': ['graph_kmer_index=graph_kmer_index.command_line_interface:main']
      },
)



import numpy as np
import logging

class BaseHashMap:
    def to_file(self, file_name):
        np.save(file_name, self._hashes)

    @classmethod
    def from_file(cls, file_name):
        data = np.load(file_name + ".npy")
        if data.dtype != np.int:
            data = data.astype(np.int)
        #map = cls([])
        #map._hashes = data
        return cls(data)

    def unhash(self, hash):
        return self._hashes[hash]


class ModuloHashMap(BaseHashMap):
    def __init__(self, hashes):
        if hashes.dtype != np.int:
            logging.info("Converting hashes to int")
            self._hashes = hashes.astype(np.int) # np.array(hashes).astype(np.int)
        else:
            self._hashes = hashes

    @classmethod
    def from_sorted_array(cls, sorted_hash_array, modulo=452930477):
        logging.info("Creating modulohashmap with modulo %s" % modulo)
        sorted_hash_array = np.unique(sorted_hash_array)
        hashes = np.zeros(modulo)
        modulo = sorted_hash_array % modulo
        hashes[modulo] = np.array(np.arange(0, len(sorted_hash_array)), dtype=np.uint32)
        logging.info("Done creating hashmap")
        return cls(hashes)

    def hash(self, key, modulo=452930477):
        #return int(self._hashes[key])
        index = self._hashes[key % modulo]
        if index == 0:
            return None

        return int(index)


class LogNHashMap(BaseHashMap):
    def __init__(self, sorted_hash_array):
        self._hashes = np.unique(sorted_hash_array)

    def hash(self, key):
        index = np.searchsorted(self._hashes, key)
        if index >= len(self._hashes) or self._hashes[index] != key:
            return None
        return index





import logging
import numpy as np
from pyfaidx import Fasta
import numba
from .read_kmers import ReadKmers

@numba.jit
def fill_zeros_from_end(array):
    for i in range(1, len(array)):
        index_pos = len(array) - i
        if array[index_pos] == 0:
            array[index_pos] = array[index_pos+1]


class ReferenceKmerIndex:
    properties = {"ref_position_to_index", "kmers", "ref_positions", "nodes"}
    def __init__(self, ref_position_to_index=None, kmers=None, ref_positions=None, nodes=None):
        self.ref_position_to_index = ref_position_to_index
        self.kmers = kmers
        self.ref_positions = ref_positions
        self.nodes = nodes

    def get_between(self, ref_start, ref_end):
        return self.kmers[
            self.ref_position_to_index[ref_start]:self.ref_position_to_index[min(len(self.ref_position_to_index)-1, ref_end)]
        ]

    def get_between_except(self, ref_start, ref_end, except_position):
        assert self.ref_positions is None
        indexes = [i for i in np.arange(ref_start, ref_end) if i != except_position]
        return self.kmers[indexes]

    def get_all_between(self, ref_start, ref_end):
        if self.ref_positions is None:
            raise Exception("This index is missing reference positions and cannot be used to get all between. "
                            "Is it made from a linear reference? If so, use get_between() instead")
        start = self.ref_position_to_index[ref_start]
        end = self.ref_position_to_index[ref_end]
        return self.kmers[start:end], self.ref_positions[start:end], self.nodes[start:end]

    @classmethod
    def from_sequence(cls, genome_sequence, k, only_store_kmers=False):
        kmers = ReadKmers.get_kmers_from_read_dynamic(genome_sequence, np.power(4, np.arange(0, k)))

        ref_position_to_index = None
        if not only_store_kmers:
            ref_position_to_index = np.arange(0, len(genome_sequence), dtype=np.uint32)
        else:
            logging.info("Only storing kmers, not index")

        if k <= 16:
            kmers = kmers.astype(np.uint32)
            logging.info("Converting kmers to 32 bit uint, since k is small enough.")
        else:
            logging.info("Kmers are stored using 64 bits")
            kmers = kmers.astype(np.uint64)

        return cls(ref_position_to_index, kmers)

    @classmethod
    def from_linear_reference(cls, fasta_file_name, reference_name="ref", k=15, only_store_kmers=False):
        logging.info("Only store kmers? %s" % only_store_kmers)
        logging.info("k=%d" % k)
        genome_sequence = str(Fasta(fasta_file_name)[reference_name])
        return cls.from_sequence(genome_sequence, k, only_store_kmers)

    @classmethod
    def from_flat_kmers(cls, flat_kmers):
        ref_positions = flat_kmers._ref_offsets
        sorting = np.argsort(ref_positions)
        ref_positions = ref_positions[sorting]
        kmers = flat_kmers._hashes[sorting]
        logging.info("Checking if kmers can be stored as 32 bit")
        if np.max(kmers) < 2**32:
            logging.warning("Storing kmers as 32 bit uint since max hash is low enough")
            kmers = kmers.astype(np.uint32)

        nodes = flat_kmers._nodes[sorting]

        assert len(kmers) < 4294967295, "Too many kmers to store (32 bit limit reached). There are %d kmers" % len(kmers)

        positions_of_new_ref_positions = np.where(np.ediff1d(ref_positions, to_begin=0))[0]
        ref_position_to_index = np.zeros(int(ref_positions[-1]) + 1, dtype=np.uint32)

        ref_position_to_index[ref_positions[positions_of_new_ref_positions]] = positions_of_new_ref_positions

        # Fill zeros in ref_position_to_index
        # if there are not kmers at every ref position, there will be some zeros (ref positions mapping nowhere)
        # we want this to map to the next ref position that has a kmer
        """
        n_corrected = 0
        for i in range(1, len(ref_position_to_index)):
            if i % 1000000 == 0:
                logging.info("Filling zeros, %d processed" % i)
            index_pos = len(ref_position_to_index) - i
            if ref_position_to_index[index_pos] == 0:
                n_corrected += 1
                ref_position_to_index[index_pos] = ref_position_to_index[index_pos+1]

        logging.info("Filled %d zeros" % n_corrected)
        """
        logging.info("Filling zeroes")
        fill_zeros_from_end(ref_position_to_index)

        return cls(ref_position_to_index, kmers, ref_positions, nodes)

    def to_file(self, file_name):

        if self.ref_position_to_index is None:
            np.savez(file_name, kmers=self.kmers)
        elif self.ref_positions is None and self.nodes is None:
            np.savez(file_name,
                 ref_position_to_index=self.ref_position_to_index,
                 kmers=self.kmers)
        else:
            np.savez(file_name,
                 ref_position_to_index=self.ref_position_to_index,
                 kmers=self.kmers,
                 ref_positions=self.ref_positions,
                 nodes=self.nodes)

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name + ".npz")
        except FileNotFoundError:
            data = np.load(file_name)

        nodes = None
        ref_positions = None
        ref_position_to_index = None

        if "nodes" in data:
            nodes = data["nodes"]
        if "ref_positions" in data:
            ref_positions = data["ref_positions"]
        if "ref_position_to_index" in data:
            ref_position_to_index = data["ref_position_to_index"]

        return cls(ref_position_to_index, data["kmers"], ref_positions, nodes)